import logging

import click

from flowtrigger.credentials.types import CredentialType, DEFAULT_CREDENTIAL_ID, GitHubCredential, GitLabCredential, \
  CredentialProvider, CredentialType_value_dict, CREDENTIAL_PROVIDER_KEY


@click.group(name='credentials')
@click.pass_context
def credentials_group(ctx: click.Context) -> None:
  if ctx.obj[CREDENTIAL_PROVIDER_KEY] is None:
    raise click.ClickException('Unable to create credentials provider')


@credentials_group.command(name='list')
@click.argument('credential_type_str', metavar='type', required=False,
                type=click.Choice([item.value for item in CredentialType]))
@click.option('--contents', 'show_contents', is_flag=True, default=False, help="Show secret contents.")
@click.pass_context
def credentials_list(ctx: click.Context, credential_type_str: str | None, show_contents: bool) -> None:
  credential_provider: CredentialProvider = ctx.obj[CREDENTIAL_PROVIDER_KEY]
  credential_type: CredentialType | None = None
  if not credential_type_str is None:
    credential_type = CredentialType_value_dict.get(credential_type_str)
    if credential_type is None:
      raise click.ClickException(f'Invalid credential type. {credential_type_str}')

  credential_list = credential_provider.get_all(credential_type)
  for credential in credential_list:
    print(f'id: {credential.credential_id}')
    print(f'  type: {credential.credential_type().value}')
    if show_contents:
      print(f'  contents: {credential.as_dict()}')
    print()


@credentials_group.command(name='set')
@click.option('--id', 'credentail_id', show_default=True, default=DEFAULT_CREDENTIAL_ID, help="Credential id")
@click.argument('credential_type_str', metavar='type', required=True,
                type=click.Choice([item.value for item in CredentialType]))
@click.argument('credential_property', metavar='property', required=True)
@click.argument('value', default=None, required=False)
@click.pass_context
def credentials_set(ctx: click.Context, credentail_id: str, credential_type_str: str, credential_property: str,
                    value: str | None) -> None:
  credential_provider: CredentialProvider = ctx.obj[CREDENTIAL_PROVIDER_KEY]

  credential_type = CredentialType_value_dict.get(credential_type_str)
  if credential_type is None:
    raise click.ClickException(f'Invalid credential type. {credential_type_str}')

  if credential_type == CredentialType.GITHUB:
    credential = credential_provider.get_github(credentail_id)
    if credential is None:
      credential = GitHubCredential.new(credentail_id)
  elif credential_type == CredentialType.GITLAB:
    credential = credential_provider.get_gitlab(credentail_id)
    if credential is None:
      credential = GitLabCredential.new(credentail_id)
  else:
    raise click.ClickException(f'Unsupported credential type. {credential_type_str}')

  if value is None:
    value = click.prompt(f'Enter value ({credential_property})')

  if not credential.set_property(credential_property, value):
    raise click.ClickException(f'Unable to set property. {credential_property}')

  if not credential_provider.save_credential(credential):
    raise click.ClickException(f'Unable to save credential. {credentail_id}')

  logging.info('Credential set.')


@credentials_group.command(name='delete')
@click.option('--id', 'credentail_id', show_default=True, default=DEFAULT_CREDENTIAL_ID, help="Credential id")
@click.argument('credential_type_str', metavar='type', required=True,
                type=click.Choice([item.value for item in CredentialType]))
@click.pass_context
def credentials_delete(ctx: click.Context, credentail_id: str, credential_type_str: str) -> None:
  credential_provider: CredentialProvider = ctx.obj[CREDENTIAL_PROVIDER_KEY]

  credential_type = CredentialType_value_dict.get(credential_type_str)
  if credential_type is None:
    raise click.ClickException(f'Invalid credential type. {credential_type_str}')

  if credential_provider.delete_credentials(credential_type, credentail_id):
    logging.info('Credential deleted.')
  else:
    logging.warning('Unable to delete credential. Does the credential exist?')
