import logging
import os
import time
from urllib.parse import urlparse, urlunparse

import click
import re

from git import Repo
from github import Github
from gitlab import Gitlab
from gitlab.exceptions import GitlabGetError
from gitlab.v4.objects import Project

from flowtrigger.credentials.types import CredentialProvider, CREDENTIAL_PROVIDER_KEY
from flowtrigger.objects import RepoReleases, Release, Fork
from flowtrigger.configuration.types import Configuration, PackageRepo, get_configuration, ReplacementType, \
  CONFIGURATION_PROVIDER_KEY
from flowtrigger.util import get_cache_dir, empty_dir

__release_url_pattern = 'https://api.github.com/repos/(?P<owner>[^/]+)/(?P<repo_name>[^/]+)/releases/(?P<release_id>\\d+)'


def __parse_release(release_url: str) -> tuple[str, str, int] | None:
  match = re.search(__release_url_pattern, release_url)
  if not match:
    return None
  logging.debug('Release url parsed. %s %s %s %d', release_url, match.group('owner'), match.group('repo_name'),
                int(match.group('release_id')))
  return match.group('owner'), match.group('repo_name'), int(match.group('release_id'))


def get_repos_with_releases(configuration: Configuration, credential_provider: CredentialProvider) -> dict[str, RepoReleases]:
  results: dict[str, RepoReleases] = {}
  g = Github(login_or_token=credential_provider.get_github().key)
  notifications = g.get_user().get_notifications(all=False, participating=False)
  for notification in notifications:
    logging.info('Notification. %s %s', notification.subject.type, notification.subject.url)
    if notification.subject.type != 'Release':
      logging.debug('Notification skipped')
      continue
    repo_parsed = __parse_release(notification.subject.url)
    if repo_parsed is None:
      logging.warning('Parsing notification url failed')
      continue
    owner, repo_name, release_id = repo_parsed
    rid = f'{owner}/{repo_name}'
    watch = configuration.get_watch(rid)
    if watch is None:
      logging.info(f'Notification skipped, repo not in watch list. %s', rid)
      continue
    elif rid in results:
      release_tags = results[rid]
    else:
      try:
        repo = g.get_repo(full_name_or_id=rid)
      except Exception as e:  # TODO: limit exception type
        logging.error(f'Notification skipped, repo not found. %s', e, exc_info=True)
        continue
      release_tags = RepoReleases(rid, repo)
      results[rid] = release_tags

    try:
      release = release_tags.repo.get_release(release_id)
    except Exception as e:  # TODO: limit exception type
      logging.error(f'Notification skipped, release not found. %d %s', release_id, e, exc_info=True)
      continue

    logging.info('Release. %d %s %s', release_id, release.html_url, release.tag_name)

    if not re.match(watch.tag_filter, release.tag_name):
      logging.info(f'Notification skipped due to tag filter. %s', watch.tag_filter)
      continue

    logging.debug('Notification release recorded')
    results[rid].add_release(release_id, release.html_url, release.tag_name, notification)

  return results


def resolve_release(repo_id: str, release_tags: RepoReleases) -> Release | None:
  if len(release_tags.releases) == 0:
    logging.info('No releases to resolve release. %s', repo_id)
    return None

  sorted_releases = sorted(release_tags.releases, key=lambda r: r.tag, reverse=True)
  release = sorted_releases[0]

  for other_release in sorted_releases[1:]:
    if other_release.notification:
      other_release.notification.mark_as_read()
      logging.debug('Other release notification marked read. %s %s', repo_id, other_release.notification.id)

  tag_git_ref = release_tags.repo.get_git_ref(f'tags/{release.tag}')

  logging.info('Release resolved. %s %s %s', repo_id, release.url, release.tag)
  return Release(repo_id, release.url, release.tag, tag_git_ref.object.sha, release.notification)


def get_project_fork(gl: Gitlab, project: Project) -> Project | None:
  fork_path = f'{gl.user.username}/{project.path}'
  logging.info('Project fork. %s', fork_path)
  try:
    project = gl.projects.get(fork_path)
    return project
  except GitlabGetError as e:
    if e.response_code != 404:
      logging.error('Unable to fetch project fork. %s %s', fork_path, e, exc_info=True)
      return None

  fork_project = project.forks.create(id=fork_path, visibility='public')
  logging.info('Project fork initialized. %s', fork_project.get_id())
  while fork_project.import_status != 'finished':
    logging.debug('Waiting for project fork completion. %s', fork_project.import_status)
    time.sleep(1.0)
    fork_project = gl.projects.get(fork_project.get_id())

  logging.info('Project fork completed')
  return fork_project


def clone_fork(credential_provider: CredentialProvider, package_repo: PackageRepo, clone_dir) -> Fork | None:
  logging.info('Cloning fork. %s %s', package_repo.owner, package_repo.repo)
  glc = credential_provider.get_gitlab()
  gl = Gitlab(private_token=glc.key)
  gl.auth()
  upstream_project = gl.projects.get(f'{package_repo.owner}/{package_repo.repo}')
  fork_project = get_project_fork(gl, upstream_project)
  if fork_project is None:
    return None

  parsed_clone_url = urlparse(fork_project.http_url_to_repo)
  clone_url = urlunparse((
    parsed_clone_url.scheme,
    f'{gl.user.username}:{glc.key}@{parsed_clone_url.netloc}',
    parsed_clone_url.path,
    parsed_clone_url.params,
    parsed_clone_url.query,
    parsed_clone_url.fragment,
  ))

  upstream_remote = 'upstream'
  logging.info('Cloning fork repo. %s %s', package_repo.branch, clone_dir)
  repo = Repo.clone_from(url=clone_url, to_path=clone_dir, branch=package_repo.branch)
  with repo.config_writer() as config:
    email = gl.user.emails.list()[0].email
    config.set_value('user', 'name', gl.user.name)
    config.set_value('user', 'email', email)
    logging.info('Clone repo user configured. %s %s', gl.user.name, email)
  upstream_remote = repo.create_remote(upstream_remote, upstream_project.http_url_to_repo)
  upstream_remote.pull(package_repo.branch, rebase='true')
  repo.delete_remote(upstream_remote)
  repo.remote('origin').push()
  logging.info('Upstream changes applied')
  return Fork(fork_project, upstream_project, repo)


def process_release(release: Release, package_repo: PackageRepo, fork: Fork) -> bool:
  release_branch = f'release/{release.tag}'
  release_title = f'Release: {release.tag}'

  logging.info('Processing release. %s %s %s %s', release.repo_id, release.tag, release_branch, release_title)

  if release_branch in [branch.name for branch in fork.repo.branches]:
    logging.debug('Checking out exisitng release branch')
    fork.repo.git.checkout(release_branch)
  else:
    logging.debug('Checking out a new release branch')
    fork.repo.git.checkout('-b', release_branch)

  for replacement in package_repo.replacements:
    match replacement.replacement_type:
      case ReplacementType.COMMIT:
        replacement_value = release.commit_id
      case ReplacementType.TAG:
        replacement_value = release.tag
      case _:
        logging.warning('Release skipped, unknown replacement type. %s', replacement.replacement_type)
        return False

    logging.info('Replacement %s %s %s', replacement.replacement_type, replacement_value, replacement.file)

    if replacement.format_exp:
      match = re.match(replacement.format_exp, replacement_value)
      if not match:
        logging.error('Release skipped, unable to format replacement value. %s', replacement.format_exp)
        return False
      replacement_value = match.group(1)

    with open(os.path.join(fork.repo.working_dir, replacement.file), 'r+', encoding='utf-8') as f:
      file_contents = f.read()
      re_replacement = replacement.replacement.replace('<value>', replacement_value)
      f.seek(0)
      updated_file_contents = re.sub(replacement.select, re_replacement, file_contents, flags=re.MULTILINE)
      f.write(updated_file_contents)
      f.truncate()

    logging.debug('Replacement file updated.')

  if fork.repo.is_dirty(untracked_files=True):
    fork.repo.git.add(all=True)
    fork.repo.index.commit(release_title)
    fork.repo.git.push('--set-upstream', 'origin', release_branch)
  else:
    logging.info('Fork repo had no changes')

  for merge_request in fork.upstream.mergerequests.list(source_branch=release_branch,
                                                        source_project_id=fork.fork.get_id()):
    if merge_request.state == 'opened':
      logging.info('Merge request already open. %s', merge_request.web_url)
      return True

  merge_request = fork.fork.mergerequests.create({
    'source_branch': release_branch,
    'target_branch': package_repo.branch,
    'target_project_id': fork.upstream.get_id(),
    'title': release_title,
    'description': f'Release: {release.url}',
    'remove_source_branch': True,
    'allow_collaboration': True
  })
  logging.info('Merge request opened. %s', merge_request.web_url)
  return True


@click.command()
@click.pass_context
def trigger(ctx: click.Context) -> None:
  configuration = get_configuration(ctx.obj[CONFIGURATION_PROVIDER_KEY])
  credential_provider: CredentialProvider = ctx.obj[CREDENTIAL_PROVIDER_KEY]
  if credential_provider is None:
    logging.error('Unable to get credential provider')
    return

  cache_dir = get_cache_dir()
  logging.debug('Cache directory. %s', cache_dir)
  empty_dir(cache_dir)

  repos_with_releases = get_repos_with_releases(configuration, credential_provider)
  logging.info('Potential repos with release. %d', len(repos_with_releases))
  for repo_id, repos_with_releases in repos_with_releases.items():
    release = resolve_release(repo_id, repos_with_releases)
    if release is None:
      continue

    package_repo = configuration.get_package_repo_by_watch_repo_id(repo_id)
    if package_repo is None:
      logging.warning('Package repo not found. %s', repo_id)
      continue
    clone_dir = os.path.join(cache_dir, package_repo.repo)
    fork = clone_fork(credential_provider, package_repo, clone_dir)
    if fork is None:
      continue
    release_processed = process_release(release, package_repo, fork)
    if release_processed and release.notification:
      release.notification.mark_as_read()
      logging.debug('Release notification marked read. %s %s', repo_id, release.notification.id)

  empty_dir(cache_dir)
