import os.path
from enum import Enum
import yaml

from flowtrigger.util import resolve_configuration_file

CONFIGURATION_FILE_NAME: str = 'configuration.yaml'

CONFIGURATION_PROVIDER_KEY = 'configuration'


class Watch:
  def __init__(self, owner_and_repo: str, package_repo: str, tag_filter: str | None):
    parts = owner_and_repo.split(sep='/', maxsplit=2)
    self.__owner: str = parts[0]
    self.__repo: str = parts[1]
    self.__package_repo = package_repo
    self.__tag_filter: str = '.+' if tag_filter is None else tag_filter

  @property
  def owner(self) -> str:
    return self.__owner

  @property
  def repo(self) -> str:
    return self.__repo

  @property
  def package_repo(self) -> str:
    return self.__package_repo

  @property
  def tag_filter(self) -> str | None:
    return self.__tag_filter


class ReplacementType(Enum):
  COMMIT = "commit"
  TAG = "tag"


class Replacement:
  def __init__(self, replacement_type: ReplacementType, file: str, select: str, replacement: str,
               format_exp: str | None):
    self.__replacement_type = replacement_type
    self.__file = file
    self.__select = select
    self.__replacement = replacement
    self.__format_exp = '(.+)' if format_exp is None else format_exp

  @property
  def replacement_type(self) -> ReplacementType:
    return self.__replacement_type

  @property
  def file(self) -> str:
    return self.__file

  @property
  def select(self) -> str:
    return self.__select

  @property
  def replacement(self) -> str:
    return self.__replacement

  @property
  def format_exp(self) -> str:
    return self.__format_exp


class PackageRepo:
  def __init__(self, owner_and_repo: str, branch: str, replacements: dict[str, list[dict[str, str]]]):
    parts = owner_and_repo.split(sep='/', maxsplit=2)
    self.__owner: str = parts[0]
    self.__repo: str = parts[1]
    self.__branch: str = branch
    self.__replacements: list[Replacement] = []

    if 'commits' in replacements:
      for replacement in replacements['commits']:
        self.__replacements.append(Replacement(
          ReplacementType.COMMIT,
          replacement['file'],
          replacement['select'],
          replacement['replacement'],
          replacement.get('format')
        ))

    if 'tags' in replacements:
      for replacement in replacements['tags']:
        self.__replacements.append(Replacement(
          ReplacementType.TAG,
          replacement['file'],
          replacement['select'],
          replacement['replacement'],
          replacement.get('format')
        ))

  @property
  def owner(self) -> str:
    return self.__owner

  @property
  def repo(self) -> str:
    return self.__repo

  @property
  def branch(self) -> str:
    return self.__branch

  @property
  def replacements(self) -> list[Replacement]:
    return self.__replacements


class Configuration:
  def __init__(self, watches: dict[str, dict], package_repos: dict[str, dict]):
    self.__watches: dict[str, Watch] = {}
    self.__package_repos: dict[str, PackageRepo] = {}

    for key, value in watches.items():
      self.__watches[key] = Watch(
        key,
        value['package_repo'],
        value.get('tag_filter')
      )

    for key, value in package_repos.items():
      self.__package_repos[key] = PackageRepo(
        key,
        value['branch'],
        value['replacements'] if 'replacements' in value else {}
      )

  @property
  def package_repos(self) -> list[PackageRepo]:
    return list(self.__package_repos.values())

  def get_watch(self, rid) -> Watch | None:
    return self.__watches.get(rid)

  def get_package_repo_by_watch_repo_id(self, watch_rid) -> PackageRepo | None:
    watch = self.get_watch(watch_rid)
    return None if watch is None else self.get_package_repo(watch.package_repo)

  def get_package_repo(self, rid) -> PackageRepo | None:
    return self.__package_repos.get(rid)


def get_configuration(path: str) -> Configuration | None:
  if path is None or not os.path.isfile(path):
    path = resolve_configuration_file(CONFIGURATION_FILE_NAME)

  with open(path, 'rt', encoding='utf-8') as f:
    data = yaml.safe_load(f.read())
  return Configuration(**data)
