import logging
import os

import yaml

from flowtrigger.credentials.types import CredentialProvider, CredentialType, Credential, GitHubCredential, GitLabCredential, \
  CredentialProviderType, DEFAULT_CREDENTIAL_ID
from flowtrigger.util import resolve_configuration_file

CREDENTIALS_FILE_NAME: str = 'credentials.yaml'


class FileCredentialProvider(CredentialProvider):
  def __init__(self, credentials: dict[str, object]):
    self.__credentials: dict[str, dict[str, Credential]] = {}

    for credential_type, credential_val in credentials.items():
      type_credentials: dict[str, Credential] = self.__credentials.get(
        credential_type) if credential_type in self.__credentials else {}
      for credential_id, credential_obj in credential_val.items():
        if credential_type == CredentialType.GITHUB.value:
          type_credentials[credential_id] = GitHubCredential(credential_id, credential_obj)
        elif credential_type == CredentialType.GITLAB.value:
          type_credentials[credential_id] = GitLabCredential(credential_id, credential_obj)
        else:
          raise ValueError(f'Unknown credentials type: {credential_type}')

      self.__credentials[credential_type] = type_credentials

  @property
  def credential_provider_type(self) -> CredentialProviderType:
    return CredentialProviderType.FILE

  def _get_by_type(self, credential_type: CredentialType, credential_id: str) -> Credential | None:
    credentials_dict = self.__credentials.get(credential_type.value)
    if credentials_dict is None:
      return None
    return credentials_dict.get(credential_id)

  def get_github(self, credential_id: str = DEFAULT_CREDENTIAL_ID) -> GitHubCredential | None:
    return self._get_by_type(CredentialType.GITHUB, credential_id)

  def get_gitlab(self, credential_id: str = DEFAULT_CREDENTIAL_ID) -> GitLabCredential | None:
    return self._get_by_type(CredentialType.GITLAB, credential_id)

  def get_all(self, credential_type: CredentialType | None = None) -> list[Credential]:
    return []

  def save_credential(self, credential: Credential) -> bool:
    logging.warning('Credential update is not supported with current credentials provider. %s',
                    self.credential_provider_type.value)
    return False

  def delete_credentials(self, credential_type: CredentialType, credential_id: str) -> bool:
    logging.warning('Credential deletion is not supported with current credentials provider. %s',
                    self.credential_provider_type.value)
    return False


def get_file_credentials(path: str | None) -> FileCredentialProvider | None:
  if path is None or not os.path.isfile(path):
    path = resolve_configuration_file(CREDENTIALS_FILE_NAME)

  with open(path, 'rt', encoding='utf-8') as f:
    data = yaml.safe_load(f.read())
  return FileCredentialProvider(**data)
