import json
import logging

from flowtrigger.credentials.types import CredentialProvider, GitHubCredential, GitLabCredential, Credential, CredentialType, \
  CredentialProviderType, DEFAULT_CREDENTIAL_ID, CredentialType_value_dict
from flowtrigger.util import PROJECT_NAME

CREDENTIAL_SCHEMA_APPLICATION_KEY = 'application'
CREDENTIAL_SCHEMA_TYPE_KEY = 'type'
CREDENTIAL_SCHEMA_ID_KEY = 'id'

__LIBSECRET_AVAILABLE: bool = False

try:
  import gi

  gi.require_version('Secret', '1')
  from gi.repository import Secret

  CREDENTIAL_SCHEMA = Secret.Schema.new(
    f'pgdev.{PROJECT_NAME}.credential',
    Secret.SchemaFlags.NONE,
    {
      CREDENTIAL_SCHEMA_APPLICATION_KEY: Secret.SchemaAttributeType.STRING,
      CREDENTIAL_SCHEMA_TYPE_KEY: Secret.SchemaAttributeType.STRING,
      CREDENTIAL_SCHEMA_ID_KEY: Secret.SchemaAttributeType.STRING
    }
  )

  __LIBSECRET_AVAILABLE = True
except Exception as e:
  logging.warning('libsecret unavailable. %s', e, exc_info=True)


def _get_secret_value(credential_type: str, credential_id: str) -> dict | None:
  secret_attributes = {
    CREDENTIAL_SCHEMA_APPLICATION_KEY: PROJECT_NAME,
    CREDENTIAL_SCHEMA_TYPE_KEY: credential_type,
    CREDENTIAL_SCHEMA_ID_KEY: credential_id
  }
  secret_value = Secret.password_lookup_sync(CREDENTIAL_SCHEMA, secret_attributes, None)
  if secret_value is None:
    return None
  return json.loads(secret_value)


class LibSecretCredentialProvider(CredentialProvider):
  @property
  def credential_provider_type(self) -> CredentialProviderType:
    return CredentialProviderType.LIBSECRET

  def get_github(self, credential_id: str = DEFAULT_CREDENTIAL_ID) -> GitHubCredential | None:
    credential_dict = _get_secret_value(CredentialType.GITHUB.value, credential_id)
    return None if credential_dict is None else GitHubCredential(credential_id, credential_dict)

  def get_gitlab(self, credential_id: str = DEFAULT_CREDENTIAL_ID) -> GitLabCredential | None:
    credential_dict = _get_secret_value(CredentialType.GITLAB.value, credential_id)
    return None if credential_dict is None else GitLabCredential(credential_id, credential_dict)

  def get_all(self, credential_type: CredentialType | None = None) -> list[Credential]:
    credentials: list[Credential] = []
    secret_search_attributes = {
      CREDENTIAL_SCHEMA_APPLICATION_KEY: PROJECT_NAME
    }
    if not credential_type is None:
      secret_search_attributes[CREDENTIAL_SCHEMA_TYPE_KEY] = credential_type.value

    credential_secret_items = Secret.password_search_sync(
      CREDENTIAL_SCHEMA,
      secret_search_attributes,
      Secret.SearchFlags.ALL,
      None
    )
    for credential_secret_item in credential_secret_items:
      secret_attributes = credential_secret_item.get_attributes()
      if not CREDENTIAL_SCHEMA_ID_KEY in secret_attributes:
        logging.warning(f'Credential does not have an id attribute')
        continue
      credential_id = secret_attributes[CREDENTIAL_SCHEMA_ID_KEY]

      if not CREDENTIAL_SCHEMA_TYPE_KEY in secret_attributes:
        logging.warning(f'Credential does not have a type attribute. %s', credential_id)
        continue
      credential_type_str = secret_attributes[CREDENTIAL_SCHEMA_TYPE_KEY]

      if not credential_type_str in CredentialType_value_dict:
        logging.warning(f'Invalid credential type. %s %s', credential_type_str, credential_id)
        continue
      credential_type = CredentialType_value_dict[credential_type_str]

      try:
        credential_secret_item.load_secret_sync()
        credential_dict = json.loads(credential_secret_item.get_secret().get_text())
        if credential_type == CredentialType.GITHUB:
          credentials.append(GitHubCredential(credential_id, credential_dict))
        elif credential_type == CredentialType.GITLAB:
          credentials.append(GitLabCredential(credential_id, credential_dict))
      except Exception as e:
        logging.error('Failed to get credential value. %s', e, exc_info=True)
        continue
    return credentials

  def save_credential(self, credential: Credential) -> bool:
    secret_attributes = {
      CREDENTIAL_SCHEMA_APPLICATION_KEY: PROJECT_NAME,
      CREDENTIAL_SCHEMA_TYPE_KEY: credential.credential_type().value,
      CREDENTIAL_SCHEMA_ID_KEY: credential.credential_id
    }
    secret_label = f'{secret_attributes[CREDENTIAL_SCHEMA_APPLICATION_KEY]}_{secret_attributes[CREDENTIAL_SCHEMA_TYPE_KEY]}_{secret_attributes[CREDENTIAL_SCHEMA_ID_KEY]}'
    secret_value = json.dumps(credential.as_dict())
    return Secret.password_store_sync(CREDENTIAL_SCHEMA, secret_attributes, Secret.COLLECTION_DEFAULT, secret_label,
                                      secret_value,
                                      None)

  def delete_credentials(self, credential_type: CredentialType, credential_id: str) -> bool:
    secret_attributes = {
      CREDENTIAL_SCHEMA_APPLICATION_KEY: PROJECT_NAME,
      CREDENTIAL_SCHEMA_TYPE_KEY: credential_type.value,
      CREDENTIAL_SCHEMA_ID_KEY: credential_id
    }

    return Secret.password_clear_sync(CREDENTIAL_SCHEMA, secret_attributes, None)


def get_libsecret_credential_provider() -> LibSecretCredentialProvider | None:
  if not __LIBSECRET_AVAILABLE:
    return None
  return LibSecretCredentialProvider()
