from abc import abstractmethod
from enum import Enum

DEFAULT_CREDENTIAL_ID = 'default'

CREDENTIAL_PROVIDER_KEY = 'credential_provider'


class CredentialProviderType(Enum):
  LIBSECRET = "libsecret"
  FILE = "file"


CredentialProviderType_value_dict = {item.value: item for item in CredentialProviderType}


class CredentialType(Enum):
  GITHUB = "github"
  GITLAB = "gitlab"


CredentialType_value_dict = {item.value: item for item in CredentialType}

GITHUB_KEY_KEY = GITLAB_KEY_KEY = 'key'
GITHUB_USERNAME_KEY = GITLAB_USERNAME_KEY = 'username'
GITHUB_PASSWORD_KEY = GITLAB_PASSWORD_KEY = 'password'
GITLAB_URL_KEY = 'url'


def set_credential_property(obj, credential_property: str, value: str) -> bool:
  if not hasattr(obj, credential_property):
    return False
  setattr(obj, credential_property, value)
  return True


class Credential:
  @classmethod
  @abstractmethod
  def credential_type(cls) -> CredentialType:
    pass

  @property
  @abstractmethod
  def credential_id(self) -> str:
    pass

  @abstractmethod
  def set_property(self, credential_property: str, value: str) -> bool:
    pass

  @abstractmethod
  def as_dict(self) -> dict[str, str]:
    pass


class GitHubCredential(Credential):
  def __init__(self, credential_id: str, d: dict[str, str]):
    self._id = credential_id
    self._key = d.get(GITHUB_KEY_KEY)
    self._username = d.get(GITHUB_USERNAME_KEY)
    self._password = d.get(GITHUB_PASSWORD_KEY)

  @classmethod
  def new(cls, credential_id: str):
    return GitHubCredential(credential_id, {})

  @classmethod
  def credential_type(cls) -> CredentialType:
    return CredentialType.GITHUB

  @property
  def credential_id(self) -> str:
    return self._id

  @property
  def key(self) -> str | None:
    return self._key

  @property
  def username(self) -> str | None:
    return self._username

  @property
  def password(self) -> str | None:
    return self._password

  def set_property(self, credential_property: str, value: str) -> bool:
    return set_credential_property(self, f'_{credential_property}', value)

  def as_dict(self) -> dict[str, str]:
    return {
      GITHUB_KEY_KEY: self.key,
      GITHUB_USERNAME_KEY: self.username,
      GITHUB_PASSWORD_KEY: self.password
    }


class GitLabCredential(Credential):
  def __init__(self, credential_id: str, d: dict[str, str]):
    self._id = credential_id
    self._key = d.get(GITLAB_KEY_KEY)
    self._username = d.get(GITLAB_USERNAME_KEY)
    self._password = d.get(GITLAB_PASSWORD_KEY)
    self._url = d.get(GITLAB_URL_KEY)

  @classmethod
  def new(cls, credential_id: str):
    return GitLabCredential(credential_id, {})

  @classmethod
  def credential_type(cls) -> CredentialType:
    return CredentialType.GITLAB

  @property
  def credential_id(self) -> str:
    return self._id

  @property
  def key(self) -> str | None:
    return self._key

  @property
  def username(self) -> str | None:
    return self._username

  @property
  def password(self) -> str | None:
    return self._password

  @property
  def url(self) -> str | None:
    return self._url

  def set_property(self, credential_property: str, value: str) -> bool:
    return set_credential_property(self, f'_{credential_property}', value)

  def as_dict(self) -> dict[str, str]:
    return {
      GITLAB_KEY_KEY: self.key,
      GITLAB_USERNAME_KEY: self.username,
      GITLAB_PASSWORD_KEY: self.password,
      GITLAB_URL_KEY: self.url
    }


class CredentialProvider:

  @property
  @abstractmethod
  def credential_provider_type(self) -> CredentialProviderType:
    pass

  @abstractmethod
  def get_github(self, credential_id: str = DEFAULT_CREDENTIAL_ID) -> GitHubCredential | None:
    pass

  @abstractmethod
  def get_gitlab(self, credential_id: str = DEFAULT_CREDENTIAL_ID) -> GitLabCredential | None:
    pass

  @abstractmethod
  def get_all(self, credential_type: CredentialType | None = None) -> list[Credential]:
    pass

  @abstractmethod
  def save_credential(self, credential: Credential) -> bool:
    pass

  @abstractmethod
  def delete_credentials(self, credential_type: CredentialType, credential_id: str) -> bool:
    pass
